/*
  Zooming pixelation effect with rotated texels
  Copyright 2000 Jetro Lauha

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/*
  Despite of feeling lazy first I decided to do something for the
  image distortion Java applet contest and here is the result, one could
  perhaps call it pixelating rotozoomer. I haven't seen this kind of
  combination before so it can almost be counted as a new effect. :-)

  The effect is limited to a fixed scale of 256 colors for speed reasons.
  I'm sure that this still isn't the fastest way to do the effect so feel
  free to think of better solutions..

  If you use this source, extend it, optimize it, fix it or have any
  comments about it, please drop me an E-mail.

    - Jetro Lauha 2000-12-21 (jetro@iki.fi)

  Uses (slightly modified) tinyptc by Gaffer (http://www.gaffer.org/tinyptc/).
*/

import java.awt.*;
import java.awt.image.*;

public class Roto extends tinyptc8
{
    // image file name
    private final String IMAGENAME = "roto.jpg";

    // dimensions for the texture
    private final int TEXTURE_WIDTH = 128;
    private final int TEXTURE_HEIGHT = 128;

    private final int FIXED_POINT = 12;
    private final int FIXED_POINT_UNIT = 1 << FIXED_POINT;

    private final int SINCOS_TABLE = 1024;

    // applet size (from HTML) MUST be equal to size specified here
    private final int WIDTH = 256;
    private final int HEIGHT = 256;
    private final int WIDTH_SH = 8;

    // result image
    private byte[] screen;
    // original image
    private int[] image;
    // texture colors (interleaved added color values and counters)
    private int[] texcol;
    // pixel to texel mapping array
    private int[] texlist;
    // sine & cosine look-up tables
    private int[] sin, cos;
    // one clear row for cleaning texture arrays
    private int[] texclr;


    /**
     * Initializes the effect data.
     */
    private void initialize()
    {
        MediaTracker mt;
        Image img;
        int a, b, r, g;
        int[] pal;

        screen = new byte[WIDTH * HEIGHT];
        image = new int[WIDTH * HEIGHT];
        texcol = new int[TEXTURE_WIDTH * TEXTURE_HEIGHT * 2];
        texlist = new int[WIDTH * HEIGHT];
        sin = new int[SINCOS_TABLE];
        cos = new int[SINCOS_TABLE];
        texclr = new int[TEXTURE_WIDTH * 2];

        for (a = 0; a < SINCOS_TABLE; ++a) {
            sin[a] = (int)(Math.sin(a * Math.PI / (SINCOS_TABLE/2)) * FIXED_POINT_UNIT);
            cos[a] = (int)(Math.cos(a * Math.PI / (SINCOS_TABLE/2)) * FIXED_POINT_UNIT);
        }

        // load the image
        img = getImage(getCodeBase(), IMAGENAME);
        mt = new MediaTracker(this);
        mt.addImage(img, 0);
        try {
            mt.waitForAll();
        } catch (Exception ex) {
        }

        // convert the image to grayscale
        int iw = img.getWidth(null), ih = img.getHeight(null);
        int[] pixels = new int[iw * ih];
        PixelGrabber pg = new PixelGrabber(img, 0, 0, iw, ih, pixels, 0, iw);
        try {
            pg.grabPixels();
        } catch (Exception ex) {
        }
        for (b = 0; b < HEIGHT; ++b) {
            for (a = 0; a < WIDTH; ++a) {
                int o = (b << WIDTH_SH) + a;
                int p = pixels[o];
                image[o] = (((p >> 16) & 0xff) +
                            ((p >> 8) & 0xff) +
                             (p & 0xff)) / 3;
            }
        }

        // create a brown tinted palette from black to white
        pal = new int[256];
        for (a = 0; a < 64; ++a) {
            int i = 0;
            r = a * 2;
            g = a;
            b = a / 2;
            pal[a + i] = (r << 16) | (g << 8) | b;
            i += 64;
            r = a + 128;
            g = a + 64;
            b = a / 2 + 32;
            pal[a + i] = (r << 16) | (g << 8) | b;
            i += 64;
            r = a + 192;
            g = a + 128;
            b = a + 64;
            pal[a + i] = (r << 16) | (g << 8) | b;
            i += 64;
            r = 255;
            g = a + 192;
            b = a * 2 + 128;
            pal[a + i] = (r << 16) | (g << 8) | b;
        }
        palette(pal);
    }


    /**
     * Draw the rotated and zoomed pixelation effect.
     * @param x X offset in texture coordinates
     * @param y Y offset in texture coordinates
     * @param rot Rotation value, must be in range [0..SINCOS_TABLE[
     * @param zoom Zoom value
     * @param cx Center x-coordinate in screen coordinates
     * @param cy Center y-coordinate in screen coordinates
     */
    private void pixelate(int x, int y, int rot, int zoom, int cx, int cy)
    {
        int xx, xy, yx, yy;
        int a, b, tx, ty, txb, tyb, c = 0, o;

        rot %= SINCOS_TABLE;

        // clear the texture colors
        for (b = 0; b < TEXTURE_HEIGHT; ++b)
            System.arraycopy(texclr, 0, texcol, b * TEXTURE_WIDTH * 2, TEXTURE_WIDTH * 2);

        // calculate the rotozoomer

        xx = (cos[rot] * zoom) >> FIXED_POINT;
        xy = (sin[rot] * zoom) >> FIXED_POINT;
        yx = (cos[(rot + SINCOS_TABLE / 4) & (SINCOS_TABLE - 1)] * zoom) >> FIXED_POINT;
        yy = (sin[(rot + SINCOS_TABLE / 4) & (SINCOS_TABLE - 1)] * zoom) >> FIXED_POINT;
        txb = x - yx * cy - xx * cx + b * yx;
        tyb = y - yy * cy - xy * cx + b * yy;

        for (b = 0; b < HEIGHT; ++b) {

            tx = txb;
            ty = tyb;
            txb += yx;
            tyb += yy;

            for (a = 0; a < WIDTH; ++a) {
                tx = (tx + xx) & (FIXED_POINT_UNIT * TEXTURE_WIDTH - 1);
                ty = (ty + xy) & (FIXED_POINT_UNIT * TEXTURE_HEIGHT - 1);

                o = ((ty >> FIXED_POINT) * TEXTURE_WIDTH + (tx >> FIXED_POINT)) << 1;
                // add the pixel color values under this texel together
                texcol[o] += image[c];
                // count of pixels under this texel
                ++texcol[o + 1];
                // map current pixel to current texel
                texlist[c++] = o;

            }
        }

        // draw the result effect
        c = 0;
        for (b = 0; b < HEIGHT; ++b) {
            for (a = 0; a < WIDTH; ++a) {
                o = texlist[c];
                if (texcol[o + 1] > 0) {
                    // calculate average color for this texel
                    texcol[o] /= texcol[o + 1];
                    texcol[o + 1] = 0;
                }
                screen[c++] = (byte)texcol[o];
            }
        }
    }


    public void main(int width, int height)
    {
        long currentTime = 0, startTime;
        int time;

        if (width != WIDTH || height != HEIGHT)
            System.err.println("Warning: Incorrect applet size!\n" +
                "Size must be " + WIDTH + " x " + HEIGHT);

        initialize();

        startTime = System.currentTimeMillis();

        int x = 0, y = 0, rot, zoom;

        showStatus(getAppletInfo());

        while (_alive) {

            // Here is a simple way to have some interpolation for the
            // timer value (on some operating systems the actual timer
            // resolution isn't good enough for smooth effects).
            // Using big constants (255,8) creates the acceleration effect.
            currentTime = (currentTime * 255 + (int)(System.currentTimeMillis() - startTime)) >> 8;

            time = (int)(currentTime / 10);

            rot = time;
            zoom = (int)(700 + 650 * sin[(time * 2) % SINCOS_TABLE] / FIXED_POINT_UNIT);

            x += 110;
            y += 90;

            pixelate(x, y, rot, -zoom, WIDTH / 2, HEIGHT / 2);

            update(screen);
        }
    }

    public String getAppletInfo()
    {
        return "Zooming pixelation effect with rotated texels\n" +
               " - Copyright 2000 Jetro Lauha (jetro@iki.fi)";
    }

}
